/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap;

import java.io.IOException;
import java.lang.reflect.Array;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;

import org.ow2.opensuit.xmlmap.impl.IXmlModelMessage;
import org.ow2.opensuit.xmlmap.impl.ObjInfo;
import org.ow2.opensuit.xmlmap.impl.ObjInfoHolder;
import org.ow2.opensuit.xmlmap.mapping.ElementMapping;
import org.ow2.opensuit.xmlmap.mapping.FieldMapping;
import org.ow2.opensuit.xmlmap.schema.ISchema;
import org.ow2.opensuit.xmlmap.schema.ISchemaElement;
import org.ow2.opensuit.xmlmap.schema.SchemasManager;
import org.ow2.opensuit.xmlmap.utils.XsdHelper;
import org.ow2.opensuit.xmlmap.utils.minidom.Parser;
import org.ow2.opensuit.xmlmap.utils.minidom.XmlAttribute;
import org.ow2.opensuit.xmlmap.utils.minidom.XmlElement;
import org.xml.sax.SAXException;



/**
 * The base class for instantiating an object model that conform to XML mapping rules.
 */
public class XmlInstantiator extends ObjInfoHolder
{
	private ClassLoader classLoader = XmlInstantiator.class.getClassLoader();
	private URL _mainDoc;
	private Object _validSupport;
	private SchemasManager _schemas;
	private Object _root;
//	private List _elements;
	
	private XmlInstantiator(URL iMainDocUrl, Object iValidSupport)
	{
		_mainDoc = iMainDocUrl;
		_validSupport = iValidSupport;
		_schemas = new SchemasManager();
//		_elements = new ArrayList();
	}
	public void setClassLoader(ClassLoader classLoader)
	{
		this.classLoader = classLoader;
		_schemas.setClassLoader(classLoader);
	}

	/**
	 * Returns the root instantiated object.
	 * @return
	 */
	public Object getRoot()
	{
		return _root;
	}
	public SchemasManager getSchemas()
	{
		return _schemas;
	}

	private void instantiateMainDoc() throws IOException, SAXException, ParserConfigurationException, FactoryConfigurationError
	{
		XmlElement elt = Parser.parse(_mainDoc);
		
		// --- 2: instanciate all
		ObjInfo docInfo = new ObjInfo(null, null, null, null);
		docInfo.setLocation(_mainDoc, 0);
		addInfo(docInfo);
		List ancestors = new ArrayList();
		ancestors.add(docInfo);

		_root = instanciateElt(ancestors, elt, _mainDoc);
		
		// --- 3: check the whole model validity
		initializeAll(_validSupport);
	}

	public final static XmlInstantiator instantiate(URL iXmlFileUrl, Object iValidSupport) throws IOException, SAXException, ParserConfigurationException, FactoryConfigurationError
	{
		return instantiate(iXmlFileUrl, iValidSupport, XmlInstantiator.class.getClassLoader());
	}
	public final static XmlInstantiator instantiate(URL iXmlFileUrl, Object iValidSupport, ClassLoader classLoader) throws IOException, SAXException, ParserConfigurationException, FactoryConfigurationError
	{
		XmlInstantiator doc = new XmlInstantiator(iXmlFileUrl, iValidSupport);
		doc.setClassLoader(classLoader);
		doc.instantiateMainDoc();
		return doc;
	}

	// ===============================================================================
	private Object instantiateImportedElt(URL fromXmlFile, String targetXmlFilePath, List iInfoAncestors) throws IOException, SAXException, ParserConfigurationException, FactoryConfigurationError
	{
		StringBuffer fileUrl = new StringBuffer();
		/*
		fileUrl.append(_mainDoc.getProtocol());
		fileUrl.append("://");
		fileUrl.append(_mainDoc.getAuthority());
		*/
		int idx = fromXmlFile.getFile().lastIndexOf('/');
		fileUrl.append(fromXmlFile.getFile().substring(0, idx+1));
		fileUrl.append(targetXmlFilePath);
		// query string? ref?
//		URL docUrl = new URL(fileUrl.toString());
		URL docUrl = new URL(fromXmlFile.getProtocol(), fromXmlFile.getHost(), fileUrl.toString());
		
		XmlElement elt = Parser.parse(docUrl);
		
		// --- 2: instanciate
		return instanciateElt(iInfoAncestors, elt, docUrl);
	}
	private boolean isSimpleType(Class type)
	{
		if(XsdHelper.getSimpleType(type) != null)
		{
			// --- this is a base XSD simple type
			return true;
		}
		ISchema declaringSchema = _schemas.getDeclaringSchema(type);
		if(declaringSchema == null)
			return false;
		if(declaringSchema.getMethod().isEnumeration(type))
			return true;
		else if(declaringSchema.getMethod().isSimpleType(type))
			return true;
		return false;
	}
	private Object instanciateSimpleType(List iInfoAncestors, Class simpleType, String stringValue, int line, URL xmlFile) throws Exception
	{
		if(XsdHelper.getSimpleType(simpleType) != null)
		{
			// --- this is a base XSD simple type
			try
			{
				return XsdHelper.parseSimpleType(stringValue, simpleType);
			}
			catch (Exception e)
			{
				throw new Exception("Could not parse base simple type '"+simpleType.getName()+"' ("+stringValue+"): "+e.getMessage(), e);
			}
		}
		ISchema declaringSchema = _schemas.getDeclaringSchema(simpleType);
		if(declaringSchema == null)
			throw new Exception("Not a simple type '"+simpleType.getName()+"'.");
		
		if(declaringSchema.getMethod().isEnumeration(simpleType))
		{
			// --- field is an enumerated type
			Object val = declaringSchema.getMethod().getEnumItem(simpleType, stringValue);
			if(val == null)
				throw new Exception("Value '"+stringValue+"' does not correspond to any element in enumeration '"+simpleType.getName()+"'.");
			return val;
		}
		else if(declaringSchema.getMethod().isSimpleType(simpleType))
		{
			ISchemaElement schemaElt = declaringSchema.getElement(simpleType);
			
			// --- compute base type value
			Object base;
			try
			{
//				base = XsdHelper.parseSimpleType(stringValue, schemaElt.getSchema().getMethod().getBaseSimpleType(simpleType));
				base = instanciateSimpleType(iInfoAncestors, schemaElt.getSchema().getMethod().getBaseSimpleType(simpleType), stringValue, line, xmlFile);
			}
			catch (Exception e)
			{
				throw new Exception("Could not get base simple type value for '"+simpleType.getName()+"': "+e.getMessage());
			}
			
			// --- instantiate object
			Object obj = null;
			try
			{
				obj = simpleType.newInstance();
			}
			catch(IllegalArgumentException e)
			{
				throw new Exception("IllegalArgumentException while invoking "+simpleType.getName()+" default constructor: "+e.getMessage());
			}
			catch(InstantiationException e)
			{
				throw new Exception("InstantiationException while invoking "+simpleType.getName()+" default constructor: "+e.getMessage());
			}
			catch(IllegalAccessException e)
			{
				throw new Exception("IllegalAccessException while invoking "+simpleType.getName()+" default constructor: "+e.getMessage());
			}
			
			// --- create info
			ObjInfo parentInfo = iInfoAncestors.size() == 0 ? null : (ObjInfo)iInfoAncestors.get(iInfoAncestors.size()-1);
			ObjInfo info = new ObjInfo(parentInfo, null, schemaElt, obj);
			info.setLocation(xmlFile, line);
			addInfo(info);
			
			// --- retrieve mappings
			ElementMapping mappings = (ElementMapping)schemaElt.getSchema().getMethod().getElementMappings(obj.getClass(), true);
			// --- log mapping errors
			for(int i=0; i<mappings.getMappingErrors().length; i++)
				info.addMessage(null, IXmlModelMessage.INST_ERROR, mappings.getMappingErrors()[i].getMessage());
			
			// =========================================================
			// === Process Root Mappings
			// =========================================================
			ObjInfo rootInfo = (ObjInfo)iInfoAncestors.get(1);
			FieldMapping[] rootMappings = mappings.getMappings(FieldMapping.ROOT);
			for(int i=0; i<rootMappings.length; i++)
			{
				FieldMapping fieldMapping = (FieldMapping)rootMappings[i];
				if(!fieldMapping.getField().getType().isInstance(rootInfo.getObject()))
				{
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Root field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the root type: "+rootInfo.getObject().getClass().getName());
				}
				else
				{
					try
					{
						fieldMapping.getField().setAccessible(true);
						fieldMapping.getField().set(obj, rootInfo.getObject());
					}
					catch(IllegalArgumentException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Root field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the root type: "+rootInfo.getObject().getClass().getName());
					}
					catch(IllegalAccessException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Root field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
					}
				}
			}
			
			// =========================================================
			// === Process Parent Mappings
			// =========================================================
	//		Object parent = iAncestors.lastElement();
			FieldMapping[] parentMappings = mappings.getMappings(FieldMapping.PARENT);
			for(int i=0; i<parentMappings.length; i++)
			{
				FieldMapping fieldMapping = (FieldMapping)parentMappings[i];
				if(!fieldMapping.getField().getType().isInstance(parentInfo.getObject()))
				{
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_WARNING, "Parent field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the parent type: "+parentInfo.getObject().getClass().getName()+". Will be null.");
				}
				else
				{
					try
					{
						fieldMapping.getField().setAccessible(true);
						fieldMapping.getField().set(obj, parentInfo.getObject());
					}
					catch(IllegalArgumentException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_WARNING, "Parent field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the parent type: "+parentInfo.getObject().getClass().getName()+". Will be null.");
					}
					catch(IllegalAccessException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Parent field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
					}
				}
			}
			
			// =========================================================
			// === Process Ancestor Mappings
			// =========================================================
			FieldMapping[] ancestorMappings = mappings.getMappings(FieldMapping.ANCESTOR);
			for(int i=0; i<ancestorMappings.length; i++)
			{
				FieldMapping fieldMapping = (FieldMapping)ancestorMappings[i];
				// --- find first ancetor that matches the type
				boolean found = false;
				for(int j=iInfoAncestors.size()-1; j>=1; j--)
				{
					ObjInfo ancestor = (ObjInfo)iInfoAncestors.get(j);
					if(fieldMapping.getField().getType().isInstance(ancestor.getObject()))
					{
						// --- type matches
						found = true;
						try
						{
							fieldMapping.getField().setAccessible(true);
							fieldMapping.getField().set(obj, ancestor.getObject());
						}
						catch(IllegalArgumentException e)
						{
							info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_WARNING, "Ancestor field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the parent type: "+ancestor.getObject().getClass().getName()+". Will be null.");
						}
						catch(IllegalAccessException e)
						{
							info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Ancestor field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
						}
						break;
					}
				}
				/*
				if(!found)
					// info
					info.logMessage(fieldMapping.getName(), IXmlModelMessage.INST_INFO, "No matching ancestor for field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+"). Will be null.");
				*/
			}
	
			// =========================================================
			// === Process Content Mapping(s)
			// =========================================================
			FieldMapping[] contentMappings = mappings.getMappings(FieldMapping.CONTENT);
			if(contentMappings.length > 0)
			{
				// --- only allow CData ?
		//		String content = XmlHelper.getContent(iNode);
				for(int i=0; i<contentMappings.length; i++)
				{
					FieldMapping fieldMapping = (FieldMapping)contentMappings[i];
					try
					{
						fieldMapping.getField().setAccessible(true);
						fieldMapping.getField().set(obj, base);
					}
					catch(IllegalArgumentException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Content field '"+fieldMapping.getField().getName()+"' is not of expected type ("+base.getClass().getName()+"): "+fieldMapping.getClass().getName());
					}
					catch(IllegalAccessException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Content field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
					}
				}
			}
			else
			{
				info.addMessage(null, IXmlModelMessage.INST_ERROR, "SimpleType object '"+simpleType.getName()+"' must declare a content mapping to get the base XSD type.");
			}
			
			// --- finally return obj
			return obj;
		}
		else
			throw new Exception("Not a simple type '"+simpleType.getName()+"'.");
	}
	private Object instanciateElt(List iInfoAncestors, XmlElement iNode, URL xmlFile)
	{
//		Object parent = iAncestors.size() == 0 ? null : (Object)iAncestors.lastElement();
		ObjInfo parentInfo = iInfoAncestors.size() == 0 ? null : (ObjInfo)iInfoAncestors.get(iInfoAncestors.size()-1);
		String nsAndTag = iNode.getTagName();
		
		// --- treat import elements
		if(iNode.getLocalName().startsWith("Import_"))
		{
			ObjInfo info = new ObjInfo(parentInfo, null, null, null);
			info.setLocation(xmlFile, iNode.getLine());
//			info.setColNumber(iNode.getCol());
			addInfo(info);

			String file = iNode.getAttribute("File");
			if(file == null)
			{
//				parentInfo.addMessage("File", IXmlModelMessage.INST_ERROR, "Attribute File expected on an Import element.");
				info.addMessage("File", IXmlModelMessage.INST_ERROR, "Attribute File expected on an Import element.");
				return null;
			}
			try
	        {
	            return instantiateImportedElt(xmlFile, file, iInfoAncestors);
	        }
	        catch(Exception e)
	        {
//	        	parentInfo.addMessage("File", IXmlModelMessage.INST_ERROR, "Error while importing document '"+file+"': "+e.toString());
	        	info.addMessage("File", IXmlModelMessage.INST_ERROR, "Error while importing document '"+file+"': "+e.toString());
	        	e.printStackTrace();
	        	return null;
	        }
		}
		
		// --- 1: get class
		Class c;
		try
		{
//			c = Class.forName(iNode.getNamespaceURI()+"."+iNode.getLocalName());
//			c = classLoader.loadClass(iNode.getNamespaceURI()+"."+iNode.getLocalName());
			c = Class.forName(iNode.getNamespaceURI()+"."+iNode.getLocalName(), true, classLoader);
		}
		catch(ClassNotFoundException e)
		{
			parentInfo.addMessage(null, IXmlModelMessage.INST_ERROR, "No class found for tag '"+nsAndTag+"'.");
			return null;
		}
		
		// --- declare schema
		if(_schemas.getSchema(iNode.getNamespaceURI()) == null)
		{
			// --- look for schema location
			String location = null;
			/*
			String locations = iNode.getOwnerDocument().getDocumentElement().getAttributeNS("http://www.w3.org/2001/XMLSchema-instance", "schemaLocation");
			if(locations != null)
			{
				StringTokenizer st = new StringTokenizer(locations, " ");
				while(st.hasMoreTokens())
				{
					String ns = st.nextToken();
					String loc = st.nextToken();
					if(ns.equals(iNode.getNamespaceURI()))
					{
						location = loc;
						break;
					}
				}
			}
			*/
			String prefix = null;
			int idx = iNode.getTagName().indexOf(':');
			if(idx > 0)
				prefix = iNode.getTagName().substring(0, idx);
			_schemas.declareSchema(prefix, location, iNode.getNamespaceURI());
		}
		
		// --- 2: instanciate (call default constructor)
		Object obj = null;
		try
		{
			Object o = c.newInstance();
			if(!(o instanceof Object))
			{
				parentInfo.addMessage(null, IXmlModelMessage.INST_ERROR, "Instancated element is not an Object: "+c);
				return null;
			}
			obj = (Object)o;
		}
		catch(IllegalArgumentException e)
		{
			parentInfo.addMessage(null, IXmlModelMessage.INST_ERROR, "IllegalArgumentException while invoking "+c.getName()+" default constructor: "+e.getMessage());
			return null;
		}
		catch(InstantiationException e)
		{
			parentInfo.addMessage(null, IXmlModelMessage.INST_ERROR, "InstantiationException while invoking "+c.getName()+" default constructor: "+e.getMessage());
			return null;
		}
		catch(IllegalAccessException e)
		{
			parentInfo.addMessage(null, IXmlModelMessage.INST_ERROR, "IllegalAccessException while invoking "+c.getName()+" default constructor: "+e.getMessage());
			return null;
		}
		/*
		catch(InvocationTargetException e)
		{
			iErrors.logMessage(iParent, INST_ERROR, "Exception while building object "+c+": "+e.getTargetException().getMessage());
			return null;
		}
		*/
		
		// --- create ObjInfo for the new object
//		createObjInfo(parent, iNode, obj);
		ISchemaElement schemaElt = _schemas.getElement(obj.getClass());
		ObjInfo info = new ObjInfo(parentInfo, null, schemaElt, obj);
		info.setLocation(xmlFile, iNode.getLine());
//		info.setColNumber(iNode.getCol());
		addInfo(info);
		
		// --- 3: now populate object
		List ancestorsForChildren = new ArrayList(iInfoAncestors);
		ancestorsForChildren.add(info);
		
		// --- extract XmlMappings
//System.out.println("schemaElt: "+schemaElt);
//System.out.println("schema: "+schemaElt.getSchema());
//System.out.println("method: "+schemaElt.getSchema().getMethod());
		ElementMapping mappings = (ElementMapping)schemaElt.getSchema().getMethod().getElementMappings(obj.getClass(), true);
		// --- log mapping errors
		for(int i=0; i<mappings.getMappingErrors().length; i++)
			info.addMessage(null, IXmlModelMessage.INST_ERROR, mappings.getMappingErrors()[i].getMessage());
		
		if(iInfoAncestors.size() > 1)
		{
			// =========================================================
			// === Process Root Mappings
			// =========================================================
//			ObjInfo rootInfo = (ObjInfo)iInfoAncestors.firstElement();
			ObjInfo rootInfo = (ObjInfo)iInfoAncestors.get(1);
			FieldMapping[] rootMappings = mappings.getMappings(FieldMapping.ROOT);
			for(int i=0; i<rootMappings.length; i++)
			{
				FieldMapping fieldMapping = (FieldMapping)rootMappings[i];
				if(!fieldMapping.getField().getType().isInstance(rootInfo.getObject()))
				{
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Root field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the root type: "+rootInfo.getObject().getClass().getName());
				}
				else
				{
					try
					{
						fieldMapping.getField().setAccessible(true);
						fieldMapping.getField().set(obj, rootInfo.getObject());
					}
					catch(IllegalArgumentException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Root field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the root type: "+rootInfo.getObject().getClass().getName());
					}
					catch(IllegalAccessException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Root field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
					}
				}
			}
			
			// =========================================================
			// === Process Parent Mappings
			// =========================================================
//			Object parent = iAncestors.lastElement();
			FieldMapping[] parentMappings = mappings.getMappings(FieldMapping.PARENT);
			for(int i=0; i<parentMappings.length; i++)
			{
				FieldMapping fieldMapping = (FieldMapping)parentMappings[i];
				if(!fieldMapping.getField().getType().isInstance(parentInfo.getObject()))
				{
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_WARNING, "Parent field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the parent type: "+parentInfo.getObject().getClass().getName()+". Will be null.");
				}
				else
				{
					try
					{
						fieldMapping.getField().setAccessible(true);
						fieldMapping.getField().set(obj, parentInfo.getObject());
					}
					catch(IllegalArgumentException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_WARNING, "Parent field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the parent type: "+parentInfo.getObject().getClass().getName()+". Will be null.");
					}
					catch(IllegalAccessException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Parent field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
					}
				}
			}
			
			// =========================================================
			// === Process Ancestor Mappings
			// =========================================================
			FieldMapping[] ancestorMappings = mappings.getMappings(FieldMapping.ANCESTOR);
			for(int i=0; i<ancestorMappings.length; i++)
			{
				FieldMapping fieldMapping = (FieldMapping)ancestorMappings[i];
				// --- find first ancetor that matches the type
				boolean found = false;
				for(int j=iInfoAncestors.size()-1; j>=1; j--)
				{
					ObjInfo ancestor = (ObjInfo)iInfoAncestors.get(j);
					if(fieldMapping.getField().getType().isInstance(ancestor.getObject()))
					{
						// --- type matches
						found = true;
						try
						{
							fieldMapping.getField().setAccessible(true);
							fieldMapping.getField().set(obj, ancestor.getObject());
						}
						catch(IllegalArgumentException e)
						{
							info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_WARNING, "Ancestor field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+") is not compatible with the parent type: "+ancestor.getObject().getClass().getName()+". Will be null.");
						}
						catch(IllegalAccessException e)
						{
							info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Ancestor field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
						}
						break;
					}
				}
				/*
				if(!found)
					// info
					info.logMessage(fieldMapping.getName(), IXmlModelMessage.INST_INFO, "No matching ancestor for field '"+fieldMapping.getField().getName()+"' type ("+fieldMapping.getBaseClass().getName()+"). Will be null.");
				*/
			}
		}
		// =========================================================
		// === Process Content Mapping(s)
		// =========================================================
		FieldMapping[] contentMappings = mappings.getMappings(FieldMapping.CONTENT);
		if(contentMappings.length > 0)
		{
			// --- only allow CData ?
//			String content = XmlHelper.getContent(iNode);
			for(int i=0; i<contentMappings.length; i++)
			{
				FieldMapping fieldMapping = (FieldMapping)contentMappings[i];
				// --- instantiate value
				Object value = null;
				try
				{
					value = instanciateSimpleType(ancestorsForChildren, fieldMapping.getField().getType(), iNode.getContent(), iNode.getLine(), xmlFile);
				}
				catch (Exception e)
				{
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Content mapping could not be parsed as simple type '"+fieldMapping.getField().getType()+"' due to: "+e.getMessage());
					continue;
				}
				try
				{
					fieldMapping.getField().setAccessible(true);
					fieldMapping.getField().set(obj, value);
				}
				catch(IllegalArgumentException e)
				{
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Content field '"+fieldMapping.getField().getName()+"' is not of expected type ("+fieldMapping.getField().getType().getName()+"): "+value);
				}
				catch(IllegalAccessException e)
				{
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Content field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
				}
			}
		}
		// =========================================================
		// === Process Attributes
		// =========================================================
		// --- 1: place attributes and children Elements in a List
		List attributes = new ArrayList();
		Collection attrs = iNode.getAttributes();
		if(attrs != null)
		{
//			for(int i=0; i<attrs.size(); i++)
			for(Iterator it = attrs.iterator(); it.hasNext(); )
			{
//				if(attrs.item(i) instanceof Attr)
				{
//					Attr attr = (Attr)attrs.item(i);
					XmlAttribute attr = (XmlAttribute)it.next();
					if("http://www.w3.org/2001/XMLSchema-instance".equals(attr.getNamespaceURI()))
//					if("schemaLocation".equals(attr.getLocalName()))
						// reserved attribute: skip
						continue;
					if(attr.getName().equals("xmlns"))
						// reserved attribute: skip
						continue;
					if(attr.getName().startsWith("xmlns:"))
						// reserved attribute: skip
						continue;
					attributes.add(attr);
				}
			}
		}
		// --- 2: set fields mapped to attributes
		FieldMapping[] attrMappings = mappings.getMappings(FieldMapping.ATTRIBUTE);
		for(int i=attrMappings.length-1; i>=0; i--)
		{
			FieldMapping fieldMapping = (FieldMapping)attrMappings[i];
			// --- look for the attribute
			XmlAttribute attr = null;
			for(int j=0; j<attributes.size(); j++)
			{
				XmlAttribute curAttr = (XmlAttribute)attributes.get(j);
				if(fieldMapping.getName().equals(curAttr.getLocalName()))
				{
					attr = curAttr;
					break;
				}
			}
//				String attrValueStr = iNode.getAttribute(attrName);
			if(attr == null)
			{
				// --- no attribute found
			}
			else
			{
				// --- attribute found
				// --- remove field
				attrMappings[i] = null;
				
				String textVal = attr.getValue();
				// --- remove from attributes to process
				attributes.remove(attr);
				
				// --- try to set the field value
				if(isSimpleType(fieldMapping.getBaseClass()))
				{
					// --- 1: instantiate
					Object val = null;
					try
					{
						val = instanciateSimpleType(ancestorsForChildren, fieldMapping.getBaseClass(), textVal, iNode.getLine(), xmlFile);
					}
					catch (Exception e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Simple type attribute '"+fieldMapping.getName()+"' could not be parsed due to: "+e.getMessage());
						continue;
					}
					// --- 2: set the field
					try
					{
						fieldMapping.getField().setAccessible(true);
						fieldMapping.getField().set(obj, val);
					}
					catch(IllegalArgumentException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Attribute '"+fieldMapping.getName()+"' was not casted to the expected type (expected: "+fieldMapping.getField().getType().getName()+" / cast: "+(val == null ? "?" : val.getClass().getName())+")");
					}
					catch(IllegalAccessException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
					}
				}
				else
				{
					// --- not a simple type !
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Field '"+fieldMapping.getField().getName()+"' is not a simple type: cannot be mapped to an attribute.");
					continue;
				}
			}
		}
		
		// --- ouput errors for all unaffected attributes
		for(int i=0; i<attributes.size(); i++)
		{
			XmlAttribute attr = (XmlAttribute)attributes.get(i);
			info.addMessage(null, IXmlModelMessage.INST_ERROR, "Attribute '"+attr.getName()+"' does not match to any field of the class '"+obj.getClass()+"'.");
		}
		// --- ouput errors for all required fields mapped to xml attributes
		for(int i=0; i<attrMappings.length; i++)
		{
			FieldMapping fieldMapping = (FieldMapping)attrMappings[i];
			if(fieldMapping == null)
				continue;
			if(fieldMapping.getMinOccurs() > 0)
				info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Attribute '"+fieldMapping.getName()+"' required.");
		}
		
		// =========================================================
		// === Process Children Elements
		// =========================================================
		// --- 1: place all children elements in a List
		List childrenElts = new ArrayList();
		List childrenList = iNode.getChildren();
		if(childrenList != null)
		{
			for(int j=0; j<childrenList.size(); j++)
			{
				if(childrenList.get(j) instanceof XmlElement)
					childrenElts.add(childrenList.get(j));
			}
		}
	
		// --- 2: set fields mapped to single nodes
		FieldMapping[] childMappings = mappings.getMappings(FieldMapping.CHILD_ELT);
		for(int i=childMappings.length-1; i>=0; i--)
		{
			FieldMapping fieldMapping = (FieldMapping)childMappings[i];
			// --- look for node
			XmlElement mappedElement = null;
			for(int j=0; j<childrenElts.size(); j++)
			{
				XmlElement curElt = (XmlElement)childrenElts.get(j);
				if(fieldMapping.getName().equals(curElt.getLocalName()))
				{
					mappedElement = curElt;
					break;
				}
			}
			
			if(mappedElement != null)
			{
				// --- set the mapping location
				info.setMappingLocation(fieldMapping.getName(), mappedElement.getLine());
				
				// --- remove from children nodes to process
				childrenElts.remove(mappedElement);
				
				// --- remove field
				childMappings[i] = null;
				
				if(isSimpleType(fieldMapping.getBaseClass()))
				{
					// --- 1: instantiate
					String textVal = mappedElement.getContent();
					Object val = null;
					try
					{
						val = instanciateSimpleType(ancestorsForChildren, fieldMapping.getBaseClass(), textVal, mappedElement.getLine(), xmlFile);
					}
					catch (Exception e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Simple type element '"+fieldMapping.getName()+"' could not be parsed due to: "+e.getMessage());
						continue;
					}
					// --- 2: set the field
					try
					{
						fieldMapping.getField().setAccessible(true);
						fieldMapping.getField().set(obj, val);
					}
					catch(IllegalArgumentException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Element '"+fieldMapping.getName()+"' was not casted to the expected type (expected: "+fieldMapping.getField().getType().getName()+" / cast: "+val.getClass().getName()+")");
					}
					catch(IllegalAccessException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
					}
					continue;
				}
				
				// --- looks like it's not a simple type...
				// --- extract child elements
				List arrayElements = new ArrayList();
				for(int k=0; k<mappedElement.getChildren().size(); k++)
				{
					if(mappedElement.getChildren().get(k) instanceof XmlElement)
						arrayElements.add(mappedElement.getChildren().get(k));
				}
				// --- check nb occurs
				if(arrayElements.size() < fieldMapping.getMinOccurs())
				{
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child node '"+fieldMapping.getName()+"' expects at least "+fieldMapping.getMinOccurs()+" children.");
				}
				else if(arrayElements.size() > fieldMapping.getMaxOccurs())
				{
					info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child node '"+fieldMapping.getName()+"' expects at most "+fieldMapping.getMaxOccurs()+" children.");
				}
				// --- set field
				if(arrayElements.size() > 0)
				{
					if(fieldMapping.getField().getType().isArray())
					{
						// --- the field allows one or more child nodes
						Class baseType = fieldMapping.getField().getType().getComponentType();
						List arrayObjects = new ArrayList();
//						Object arr = Array.newInstance(baseType, arrayElements.size());
						for(int j=0; j<arrayElements.size(); j++)
						{
							XmlElement elt = (XmlElement)arrayElements.get(j);
							Object o = instanciateElt(ancestorsForChildren, elt, xmlFile);
							if(o != null)
							{
								if(!baseType.isInstance(o))
								{
									info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child element is not of expected type: '"+baseType.getName()+"'.");
								}
								else
								{
//									Array.set(arr, j, o);
									arrayObjects.add(o);
								}
							}
						}
						// --- create the array
						Object arr = Array.newInstance(baseType, arrayObjects.size());
						for(int j=0; j<arrayObjects.size(); j++)
							Array.set(arr, j, arrayObjects.get(j));
					
						// --- 2: set the field
						try
						{
							fieldMapping.getField().setAccessible(true);
							fieldMapping.getField().set(obj, arr);
						}
						catch(IllegalArgumentException e)
						{
							info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Instanciated elements array '"+fieldMapping.getName()+"' is not of the expected type (expected: "+baseType.getName()+"[] / instanciated: "+arr.getClass().getComponentType().getName()+"[])");
						}
						catch(IllegalAccessException e)
						{
							info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
						}
					}
					else
					{
						// --- the field allows only one child node
						if(arrayElements.size() > 1)
							info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Only one '"+fieldMapping.getName()+"' child node expected.");
						// --- instanciate anyway
						XmlElement elt = (XmlElement)arrayElements.get(0);
						Object o = instanciateElt(ancestorsForChildren, elt, xmlFile);
						if(o != null)
						{
							if(!fieldMapping.getField().getType().isInstance(o))
							{
								info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child element is not of expected type: '"+fieldMapping.getField().getType().getName()+"'.");
							}
							else
							{
								// --- set the field
								try
								{
									fieldMapping.getField().setAccessible(true);
									fieldMapping.getField().set(obj, o);
								}
								catch(IllegalArgumentException e)
								{
									info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Instanciated element '"+fieldMapping.getName()+"' is not of the expected type (expected: "+fieldMapping.getField().getType().getName()+" / instanciated: "+o.getClass().getName()+")");
								}
								catch(IllegalAccessException e)
								{
									info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
								}
							}
						}
					}
				}
			}
		}

		// --- 3: ouput errors for all fields mapped to nodes with minOccurs > 0
		for(int i=0; i<childMappings.length; i++)
		{
			FieldMapping fieldMapping = (FieldMapping)childMappings[i];
			if(fieldMapping == null)
				continue;
			if(fieldMapping.getMinOccurs() > 0)
				info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child node '"+fieldMapping.getName()+"' is required for class '"+obj.getClass()+"'.");
		}
		
		// =========================================================
		// === Process Collected Elements
		// =========================================================
		List collectableObjs = new ArrayList();
		// --- 1: Instantiate all remaining children nodes
		for(int i=childrenElts.size()-1; i>=0; i--)
		{
			XmlElement curElt = (XmlElement)childrenElts.get(i);
			Object o = instanciateElt(ancestorsForChildren, curElt, xmlFile);
			if(o == null)
				childrenElts.remove(i);
				// (no error message: already thrown by instantiate)
			else
				collectableObjs.add(0, o);
		}
		// --- 2: affect objects to collect fields
		FieldMapping[] collectedMappings = mappings.getMappings(FieldMapping.DIRECT_CHILD_ELT);
		for(int i=collectedMappings.length-1; i>=0; i--)
		{
			FieldMapping fieldMapping = (FieldMapping)collectedMappings[i];
			// --- this field is implicitly mapped nodes of the required type
			// --- 1: collect all objects of the according class
			Class baseType = fieldMapping.getBaseClass();
			List matchingObjects = new ArrayList();
			List matchingElements = new ArrayList();
			for(int j=collectableObjs.size()-1; j>=0; j--)
			{
				Object o = collectableObjs.get(j);
				if(baseType.isInstance(o))
				{
					collectableObjs.remove(j);
					Object elt = childrenElts.remove(j);
					matchingObjects.add(0, o);
					matchingElements.add(0, elt);
				}
			}
			
			if(matchingElements.size() > 0)
			{
				// --- set mapping location (on first item)
				info.setMappingLocation(fieldMapping.getName(), ((XmlElement)matchingElements.get(0)).getLine());
			}
			
			// --- check nb occurs
			if(matchingObjects.size() < fieldMapping.getMinOccurs())
			{
				info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child node '"+fieldMapping.getName()+"' expects at least "+fieldMapping.getMinOccurs()+" children.");
			}
			else if(matchingObjects.size() > fieldMapping.getMaxOccurs())
			{
				info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Child node '"+fieldMapping.getName()+"' expects at most "+fieldMapping.getMaxOccurs()+" children.");
			}
			// --- set field
			if(matchingObjects.size() > 0)
			{
				if(fieldMapping.getField().getType().isArray())
				{
					// --- the field allows one or more child nodes
					Object arr = Array.newInstance(baseType, matchingObjects.size());
					for(int j=0; j<matchingObjects.size(); j++)
						Array.set(arr, j, matchingObjects.get(j));
					// --- 2: set the field
					try
					{
						fieldMapping.getField().setAccessible(true);
						fieldMapping.getField().set(obj, arr);
					}
					catch(IllegalArgumentException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Instanciated elements array '"+fieldMapping.getName()+"' is not of the expected type (expected: "+baseType.getName()+"[] / instanciated: "+arr.getClass().getComponentType().getName()+"[])");
					}
					catch(IllegalAccessException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
					}
				}
				else
				{
					// --- the field allows only one child node
					if(matchingObjects.size() > 1)
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Only one '"+fieldMapping.getName()+"' child node expected.");
					// --- instanciate anyway
					// --- set the field
					Object o = matchingObjects.get(0);
					try
					{
						fieldMapping.getField().setAccessible(true);
						fieldMapping.getField().set(obj, o);
					}
					catch(IllegalArgumentException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Instanciated element '"+fieldMapping.getName()+"' is not of the expected type (expected: "+fieldMapping.getField().getType().getName()+" / instanciated: "+o.getClass().getName()+")");
					}
					catch(IllegalAccessException e)
					{
						info.addMessage(fieldMapping.getName(), IXmlModelMessage.INST_ERROR, "Field '"+fieldMapping.getField().getName()+"' cannot be set in class '"+obj.getClass().getName()+"'.");
					}
				}
			}
		}
		// --- 4: log error for all children elements not affected
		for(int i=0; i<childrenElts.size(); i++)
		{
			XmlElement elt = (XmlElement)childrenElts.get(i);
			Object o = collectableObjs.get(i);
			info.addMessage(null, IXmlModelMessage.INST_ERROR, "Node '"+elt.getTagName()+"' (--> "+o.getClass()+") does not match to any field of the class '"+obj.getClass()+"'.");
		}
		
		return obj;
	}

}
