/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap.impl;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.ow2.opensuit.xmlmap.mapping.ElementMapping;
import org.ow2.opensuit.xmlmap.mapping.FieldMapping;
import org.ow2.opensuit.xmlmap.mapping.IMappingMethod;
import org.ow2.opensuit.xmlmap.schema.ISchema;
import org.ow2.opensuit.xmlmap.schema.ISchemaElement;
import org.ow2.opensuit.xmlmap.schema.SchemasManager;
import org.ow2.opensuit.xmlmap.utils.XsdHelper;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.Element;



/**
 * This class is a program (main) that allows generating the XML schema (XSD)
 * from Java code that support the implicit mapping (see XmlObject.java) 
 */
public class SchemaImpl implements ISchema
{
	private static ISchema[] SCHEMAS_ARRAY_TYPE = {};
	private static ISchemaElement[] ELTS_ARRAY_TYPE = {};
	private static SchemaElement NO_ELT = new SchemaElement();
	
	private SchemasManager _manager;
	private List _usedSchemas = new ArrayList();
	private boolean _manageInheritance = true;
	private boolean _built = false;
	
	private String _name;
	private String _locator;
	private String _package;
	private IMappingMethod _method;
	private Hashtable _class2Infos = new Hashtable();
	
	/**
	 * Schema constructor.
	 * @param iName The schema name (generally used as the filename and the prefix).
	 * @param iPackage The root code package
	 */
	public SchemaImpl(SchemasManager iMgr, String iName, String iLocator, String iPackage)
	{
		_manager = iMgr;
		_name = iName;
		_locator = iLocator;
		_package = iPackage;
	}
	/**
	 * Returns the schema name.
	 * @return
	 */
	public String getName()
	{
		return _name;
	}
	public String getLocator()
	{
	    return _locator;
	}
	private IMappingMethod instantiateMappingMethod(String methodClassName) throws Exception
	{
		try
		{
//			Class methodClass = Class.forName(methodClassName);
			Class methodClass = Class.forName(methodClassName, true, _manager.getClassLoader());
			if(!IMappingMethod.class.isAssignableFrom(methodClass))
				throw new Exception("Mapping Method class '"+methodClassName+"' is not a valid mapping method implementation.");

			return (IMappingMethod)methodClass.newInstance();
		}
		catch (ClassNotFoundException e)
		{
			throw new Exception("Mapping Method class '"+methodClassName+"' not found.", e);
		}
		catch (InstantiationException e)
		{
			throw new Exception("Error while instantiating Mapping Method class '"+methodClassName+"'.", e);
		}
		catch (IllegalAccessException e)
		{
			throw new Exception("Could not instantiate Mapping Method class '"+methodClassName+"'.", e);
		}
	}
	private void info(String message)
	{
		System.out.println("[xmlmap:"+_package+"] INFO: "+message);
	}
	private void error(String message, Throwable t)
	{
	System.err.println("[xmlmap:"+_package+"] ERROR: "+message);
		if(t != null)
			t.printStackTrace(System.err);
	}
	public IMappingMethod getMethod()
	{
		if(_method == null)
		{
			// --- find binding method for this schema
			
			// --- 1: try the explicit way (look for a xmlmap.properties file in the root package)
			InputStream configIn = getClass().getClassLoader().getResourceAsStream((_package).replace('.', '/')+"/xmlmap.properties");
			if(configIn != null)
			{
				try
				{
					Properties config = new Properties();
					config.load(configIn);
					String methodClassName = config.getProperty("method");
					if(methodClassName != null)
					{
						try
						{
							_method = instantiateMappingMethod(methodClassName);
							info("Use Mapping Method '"+methodClassName+"'");
						}
						catch (Exception e)
						{
							error(e.getMessage(), e);
						}
					}
				}
				catch (IOException e)
				{
					error("Error while loading xmlmap configuration from "+(_package).replace('.', '/')+"/xmlmap.properties", e);
				}
			}
			
			// --- 2: try default methods
			// --- Annotations Method
			if(_method == null)
			{
				try
				{
					_method = instantiateMappingMethod("org.ow2.opensuit.xmlmap.BindWithAnnotations");
					info("Selected default mapping method '"+_method.getClass().getName()+"'");
				}
				catch (Exception e)
				{
				}
			}
			// --- Naming Method
			if(_method == null)
			{
				try
				{
					_method = instantiateMappingMethod("org.ow2.opensuit.xmlmap.BindWithNamingRules");
					info("Selected default mapping method '"+_method.getClass().getName()+"'");
				}
				catch (Exception e)
				{
				}
			}
			// --- Implicit Method
			if(_method == null)
			{
				try
				{
					_method = instantiateMappingMethod("org.ow2.opensuit.xmlmap.BindImplicitely");
					info("Selected default mapping method '"+_method.getClass().getName()+"'");
				}
				catch (Exception e)
				{
				}
			}
			
			if(_method == null)
				error("No mapping method found. Please add an xmlmap-method jar file to your classpath.", null);
		}
		return _method;
	}
	/**
	 * Returns the schema root code package
	 * @return
	 */
	public String getRootPackage()
	{
		return _package;
	}
	public boolean isFullyBuilt()
	{
	    return _built;
	}
	public String toString()
	{
		return "Schema["+_name+","+_locator+","+_package+"]";
	}
	private void setUsedSchema(ISchema iSchema)
	{
		if(iSchema == this)
			return;
		if(!_usedSchemas.contains(iSchema))
			_usedSchemas.add(iSchema);
	}
	/**
	 * 
	 * @param iSourceRootDir
	 * @return
	 */
	public ISchemaElement getElement(Class iClass)
	{
		SchemaElement elt = (SchemaElement)_class2Infos.get(iClass);
		if(elt == null)
		{
			/* No: Java 1.5
			if(iClass.isAnonymousClass())
				return null;
			*/
			
			String className = iClass.getName();
			if(!className.startsWith(_package+"."))
				return null;
			/*
			ISchema schema = _manager.getDeclaringSchema(iClass);
			if(schema != this)
				// --- class does not belong to this schema: redirect
				return schema.getElement(iClass);
			*/
			
			String tag = className.substring(_package.length()+1);
			
//			trace("Generate XSD for class "+iClass.getName()+"...");
			if(iClass.isInterface())
			{
				if(getMethod().isSubstitutionGroup(iClass))
				{
					// --- treat interface as a substitution group
					elt = new SchemaElement();
					elt._type = ISchemaElement.SUBSTITUTION_GROUP_TYPE;
					elt._class = iClass;
					elt._tag = tag;
					elt._schema = this;
				}
				else
				{
					// --- ignore this interface
					elt = NO_ELT;
				}
			}
			else if(getMethod().isEnumeration(iClass))
			{
				// --- declare enumerated type
				elt = new SchemaElement();
				elt._type = ISchemaElement.ENUMERATION_TYPE;
				elt._class = iClass;
				elt._tag = tag;
				elt._schema = this;
			}
			else if(getMethod().isSimpleType(iClass))
			{
				// --- declare enumerated type
				elt = new SchemaElement();
				elt._type = ISchemaElement.SIMPLE_TYPE_TYPE;
				elt._class = iClass;
				elt._tag = tag;
				elt._schema = this;
			}
			else if(!getMethod().isXmlElement(iClass))
			{
				// --- this class has no Xml Mapping rule: skip
//				trace(" --> no Xml mapping: skip.");
				elt = NO_ELT;
			}
			else
			{
				// --- element type
				elt = new SchemaElement();
				elt._type = ISchemaElement.ELEMENT_TYPE;
				elt._class = iClass;
				elt._tag = tag;
				elt._schema = this;
				
				// TODO: ignore class if abstract and no mapping?
				
	//			boolean manageInheritance = (_settings & MANAGE_INHERITANCE) != 0;
	//			boolean isAbstract = (iClass.getModifiers() & Modifier.ABSTRACT) != 0;
				
				// --- retrieve all implemented interfaces (for substitution groups)
				Class cls = iClass;
				List substGroupElements = new ArrayList();
				List extendedElements = new ArrayList();
				while (cls != Object.class && cls != null)
				{
					// --- (+) treat abstract classes as interfaces (this class belongs to the substitution group)
					if(cls != iClass && (cls.getModifiers() & Modifier.ABSTRACT) != 0)
					{
						// --- consider superclass as an interface (substitution group)
						// --- check that superclass is found in imports
						ISchemaElement abstractClassElement = _manager.getElement(cls);
						if(abstractClassElement != null)
						{
			                // --- interface is in a declared namespace: it is mapped to a substitution group
							extendedElements.add(abstractClassElement);
							setUsedSchema(abstractClassElement.getSchema());
						}
					}
					// --- (-)
					
					Class[] itfs = cls.getInterfaces();
					for(int i=0; i<itfs.length; i++)
					{
						/*
						if(itfs[i] == IXmlObject.class)
							// --- ignore
							continue;
						*/
						if(!getMethod().isSubstitutionGroup(itfs[i]))
							// --- ignore
							continue;
						
						// --- check that interface is found in imports
						ISchemaElement itfElement = _manager.getElement(itfs[i]);
						if(itfElement != null && !substGroupElements.contains(itfElement))
						{
			                // --- interface is in a declared namespace: it is mapped to a substitution group
							substGroupElements.add(itfElement);
							setUsedSchema(itfElement.getSchema());
						}
					}
					cls = cls.getSuperclass();
				}
				if(substGroupElements.size() > 0)
				{
					elt._substitutionGroups = (ISchemaElement[])substGroupElements.toArray(ELTS_ARRAY_TYPE);
				}
				else if(extendedElements.size() > 0)
				{
					elt._substitutionGroups = (ISchemaElement[])extendedElements.toArray(ELTS_ARRAY_TYPE);
				}
				
				// --- retrieve superclass
				Class superClass = iClass.getSuperclass();
				if(_manageInheritance && superClass != Object.class)
				{
					// --- look for super class in imports
					elt._superClass = _manager.getElement(superClass);
					if(elt._superClass != null)
						setUsedSchema(elt._superClass.getSchema());
				}
				
				// --- compute xml mappings for the class
//				elt._mappings = XmlMappings.getMappings(iClass, _manageInheritance);
			}
			// --- cache schema element
			_class2Infos.put(iClass, elt);
		}
		
		if(elt == NO_ELT)
			return null;
		return elt;
	}
	/*
	private Class[] getSubstitutionGroupInterfaces(Class iClass, Class iExcludeFrom)
	{
		Class cls = iClass;
		Vector v = new Vector();
		while (cls != iExcludeFrom && cls != null)
		{
			Class[] itfs = cls.getInterfaces();
			for(int i=0; i<itfs.length; i++)
			{
				if(itfs[i] == IXmlObject.class)
					// --- ignore
					continue;
				
				// --- check that interface is found in imports
				if(getDeclaringSchema(itfs[i]) != null)
	                // --- interface is in a declared namespace: it is mapped to a substitution group
					v.add(itfs[i]);
			}
			cls = cls.getSuperclass();
		}
		Class[] ret = new Class[v.size()];
		v.copyInto(ret);
		return ret;
	}
	*/
	
	/*
	private ISchema getDeclaringSchema(Class iClass)
	{
		String classname = iClass.getName();
		int lastDotIdx = classname.lastIndexOf('.');
		String pack = classname.substring(0, lastDotIdx);
		
		if(pack.startsWith(getRootPackage()))
		{
			// --- element belongs to this schema
			return this;
		}
		
		// --- look in imports
		if(_references != null)
		{
			for(int i=0; i<_references.size(); i++)
			{
				ISchema schema = (ISchema)_references.elementAt(i);
//				if(imp._package.equals(pack))
				if(pack.startsWith(schema.getRootPackage()))
					return schema;
			}
		}
		return null;
	}
	*/

	public ISchema[] getUsedSchemas()
	{
		return (ISchema[])_usedSchemas.toArray(SCHEMAS_ARRAY_TYPE);
	}
	public ISchemaElement[] getAllElements()
	{
		// --- remove NO_ELTS from return
		ArrayList ret = new ArrayList();
		Enumeration elements = _class2Infos.elements();
		while(elements.hasMoreElements())
		{
			ISchemaElement elt = (ISchemaElement)elements.nextElement();
			if(elt != NO_ELT)
				ret.add(elt);
		}
		
	    return (ISchemaElement[])ret.toArray(ELTS_ARRAY_TYPE);
	}
	public void buildFromClassPath()
	{
		String packageEntryName = _package.replace('.', '/')+"/";
		System.out.println("Building Schema for package: "+_package+"...");
    	Enumeration res = null;
        try
        {
	        res = getClass().getClassLoader().getResources(packageEntryName);
        }
        catch(IOException e)
        {
        }
		if(res == null)
		{
			System.out.println(" --> Directory "+packageEntryName+" not found in classpath.");
			return;
		}
		
		while(res.hasMoreElements())
		{
			URL url = (URL)res.nextElement();
			if(url.getProtocol().equals("file"))
			{
				// --- Url of type: file:<directoy_path>
				File packageDir = new File(URLDecoder.decode(url.getFile()));
				if(packageDir.isDirectory())
				{
					System.out.println(" --> Directory found: "+packageDir);
					int elts = buildFromFiles(packageDir, _package, true);
					System.out.println("     ("+elts+" elements)");
				}
			}
			else if(url.getProtocol().equals("jar") || url.getProtocol().equals("zip"))
			{
				// --- Url of type: jar:<jar_url>!<package_zip_entry>
				String jarUrlStr = URLDecoder.decode(url.getFile());
				int idx = jarUrlStr.lastIndexOf('!');
				if(idx > 0) jarUrlStr = jarUrlStr.substring(0, idx);
				System.out.println(" --> Jar found: "+jarUrlStr);
				int elts = 0;
				try
                {
	                URL jarUrl = new URL(jarUrlStr);
					ZipInputStream zip = new ZipInputStream(jarUrl.openStream());
					ZipEntry entry = null;
					while((entry = zip.getNextEntry()) != null)
					{
                    	if(!entry.isDirectory() && entry.getName().startsWith(packageEntryName) && entry.getName().endsWith(".class"))
                    	{
                    		// --- add element
    						String className = entry.getName().substring(0, entry.getName().length() - 6).replace('/', '.');
    						try
    						{
//    							Class c = Class.forName(className);
//    							Class c = _manager.getClassLoader().loadClass(className);
    							Class c = Class.forName(className, false, _manager.getClassLoader());
    							if(getElement(c) != null)
    								elts++;
    						}
    						catch(Throwable e)
    						{
    							System.out.println("Error while computing schema element for class '"+className+"'.");
    							e.printStackTrace();
    						}
                    	}
    					zip.closeEntry();
					}
					zip.close();
					System.out.println("     ("+elts+" elements)");
                }
                catch(MalformedURLException e1)
                {
	                e1.printStackTrace();
                }
                catch(IOException e)
                {
	                e.printStackTrace();
                }
			}
			else
			{
				System.out.println(" --> Unsupported protocol: "+url);
			}
		}
	}
	/**
	 * Parses recursively all classes in the given source directory
	 * @param iSourceRootDir
	 * @return
	 * @throws Exception 
	 */
	/*
	public void buildFromSource(String iSourceDir) throws Exception
	{
		File packageDir = new File(iSourceDir+"/"+(_package.replace('.', '/')));
		if(!packageDir.exists())
			throw new Exception("Package dir "+packageDir+" not found.");
		
		buildFromFiles(packageDir, _package, false);
		_built = true;
	}
	*/
	private int buildFromFiles(File packageDir, String iPackage, boolean iSearchClassFiles)
	{
//			trace("Processing package "+iPackage+"...");
		if(!packageDir.exists())
			return 0;
		
		String[] names = packageDir.list();
		int processed = 0;
		for(int i=0; i<names.length; i++)
		{
			File f = new File(packageDir, names[i]);
			if(f.exists())
			{
				if(f.isDirectory())
				{
					// --- recursive
					processed += buildFromFiles(new File(packageDir, names[i]), iPackage+"."+names[i], iSearchClassFiles);
				}
				else
				{
					if(names[i].endsWith(iSearchClassFiles ? ".class" : ".java"))
					{
						// --- source file. Get class and process.
						String className = iPackage+"."+names[i].substring(0, names[i].length() - (iSearchClassFiles ? 6 : 5));
						try
						{
//							Class c = Class.forName(className);
//							Class c = _manager.getClassLoader().loadClass(className);
							Class c = Class.forName(className, false, _manager.getClassLoader());
							if(getElement(c) != null)
								processed++;
						}
						catch(Throwable e)
						{
							System.out.println("Error while computing schema element for class '"+className+"'.");
							e.printStackTrace();
						}
					}
				}
			}
		}
		return processed;
	}
	public void generateXSD(Document iXmlDoc)
	{
		ISchemaElement[] elements = getAllElements();
		
		// --- create root element (schema)
		// --------------------------------
		Element schemaElt = iXmlDoc.createElement("xs:schema");
		schemaElt.setAttribute("xmlns:xs", "http://www.w3.org/2001/XMLSchema");
		schemaElt.setAttribute("xmlns", getRootPackage());
		schemaElt.setAttribute("targetNamespace", getRootPackage());
		
		// --- process import(s)
		// ---------------------
		ISchema[] imports = getUsedSchemas();
		for(int i=0; i<imports.length; i++)
		{
			schemaElt.setAttribute("xmlns:"+imports[i].getName(), imports[i].getRootPackage());
			
			Element importNode = iXmlDoc.createElement("xs:import");
			importNode.setAttribute("schemaLocation", imports[i].getLocator());
			importNode.setAttribute("namespace", imports[i].getRootPackage());
			schemaElt.appendChild(importNode);
		}
		iXmlDoc.appendChild(schemaElt);
		
		// --- declare base type
		// --- import type
		/*
		  <xs:complexType name="_page.ImportPage">
		    <xs:attribute name="File" type="xs:string" use="required"/>
		  </xs:complexType>
		  <xs:element name="page.ImportPage" substitutionGroup="page.IPage" type="_page.ImportPage"/>
		 */
		Element importTypeElt = iXmlDoc.createElement("xs:complexType");
		importTypeElt.setAttribute("name", "__import");
		Element importTypeAttr = iXmlDoc.createElement("xs:attribute");
		importTypeAttr.setAttribute("name", "File");
		importTypeAttr.setAttribute("type", "xs:string");
		importTypeAttr.setAttribute("use", "required");
		importTypeElt.appendChild(importTypeAttr);
		schemaElt.appendChild(importTypeElt);

		// --- declare simple types
		// ----------------------------
		System.out.println("Declaring simple types...");
		for(int i=0; i<elements.length; i++)
		{
			if(elements[i].getType() == ISchemaElement.SIMPLE_TYPE_TYPE)
			{
				Class baseType = getMethod().getBaseSimpleType(elements[i].getMappedClass());
				String baseTypeXsd = XsdHelper.getSimpleType(baseType);
				if(baseTypeXsd == null)
				{
					System.out.println(" --> warning: class '"+elements[i].getMappedClass().getName()+"' has non-simple base type ("+baseType.getName()+")");
					continue;
				}
				System.out.println(" --> simple type '"+elements[i].getTagName()+"' restricts: "+baseTypeXsd);
				
				Element typeElt = iXmlDoc.createElement("xs:simpleType");
				typeElt.setAttribute("name", elements[i].getTagName());
				addComponentDoc(typeElt, elements[i]);
				
				Element restrictionElt = iXmlDoc.createElement("xs:restriction");
				restrictionElt.setAttribute("base", "xs:"+baseTypeXsd);
				
				// TODO: add declared restrictions ???
				
				typeElt.appendChild(restrictionElt);
				schemaElt.appendChild(typeElt);
			}
		}
		
		// --- declare enumerated types
		// ----------------------------
		System.out.println("Declaring enumerated types...");
		for(int i=0; i<elements.length; i++)
		{
			if(elements[i].getType() == ISchemaElement.ENUMERATION_TYPE)
			{
				int nbElts = getMethod().getNbOfEnumItems(elements[i].getMappedClass());
				System.out.println(" --> enumeration '"+elements[i].getTagName()+"': "+nbElts+" items");
				
				Element typeElt = iXmlDoc.createElement("xs:simpleType");
				typeElt.setAttribute("name", elements[i].getTagName());
				addComponentDoc(typeElt, elements[i]);
				
				Element restrictionElt = iXmlDoc.createElement("xs:restriction");
				restrictionElt.setAttribute("base", "xs:string");
				
				for(int j=0; j<nbElts; j++)
				{
					Element enumElt = iXmlDoc.createElement("xs:enumeration");
					String enumName = getMethod().getEnumItemName(elements[i].getMappedClass(), j);
					enumElt.setAttribute("value", enumName);
					restrictionElt.appendChild(enumElt);
					addEnumDoc(enumElt, elements[i], enumName);
				}
				
				typeElt.appendChild(restrictionElt);
				schemaElt.appendChild(typeElt);
			}
		}
		
		// --- declare substitution groups
		// -------------------------------
		System.out.println("Declaring substitution groups...");
		for(int i=0; i<elements.length; i++)
		{
			if(elements[i].getType() == ISchemaElement.SUBSTITUTION_GROUP_TYPE)
			{
				System.out.println(" --> interface '"+elements[i].getTagName()+"'");
				Element ifElt = iXmlDoc.createElement("xs:element");
				ifElt.setAttribute("name", elements[i].getTagName());
				ifElt.setAttribute("abstract", "true");
				schemaElt.appendChild(ifElt);
				addComponentDoc(ifElt, elements[i]);
				
				// --- add import element if requested
				if(getMethod().isImportable(elements[i].getMappedClass()))
				{
					// --- create an import element
					Element importElt = iXmlDoc.createElement("xs:element");
					String importEltName = "Import_"+elements[i].getTagName();
					/*
					int idx = elements[i].getTagName().lastIndexOf('.');
					if(idx > 0)
						importEltName = elements[i].getTagName().substring(0, idx+1)+"Import"+elements[i].getTagName().substring(idx+1);
					*/
					
					importElt.setAttribute("name", importEltName);
					importElt.setAttribute("substitutionGroup", elements[i].getTagName());
					importElt.setAttribute("type", "__import");
					schemaElt.appendChild(importElt);
				}
			}
		}

		// --- declare elements
		// --------------------
		System.out.println("Declaring elements and types...");
		for(int i=0; i<elements.length; i++)
		{
			if(elements[i].getType() == ISchemaElement.ELEMENT_TYPE)
			{
				System.out.println(" --> element '"+elements[i].getTagName()+"' and type '_"+elements[i].getTagName()+"'");
				boolean isAbstract = (elements[i].getMappedClass().getModifiers() & Modifier.ABSTRACT) != 0;
				/*
				boolean manageInheritance = (iSettings & MANAGE_INHERITANCE) != 0;
				// --- retrieve the main implemented interface (for substitution group)
				Class[] interfaces = getSubstitutionGroupInterfaces(iClass, Object.class);
				if(interfaces.length > 1)
					throw new Exception("More than one interfaces implemented by '"+iClass.getName()+"'.");
				Class mainInterface = interfaces.length == 0 ? null : interfaces[0];
				// --- retrieve superclass
				Class superClass = iClass.getSuperclass();
				// --- retrieve xml mappings from the class
				XmlMappings mappings = XmlMappings.getMappings(iClass, manageInheritance ? superClass : Object.class);
				*/
				// --- log mapping errors
//				XmlMappings mappings = elements[i].getClassMappings();
				ElementMapping mappings = (ElementMapping)getMethod().getElementMappings(elements[i].getMappedClass(), !_manageInheritance);
				for(int j=0; j<mappings.getMappingErrors().length; j++)
					System.err.println(mappings.getMappingErrors()[j].getMessage());
				
				// --- retrieve mappings by type
				FieldMapping[] contentMappings = mappings.getMappings(FieldMapping.CONTENT);
				FieldMapping[] childMappings = mappings.getMappings(FieldMapping.CHILD_ELT);
				FieldMapping[] collectedMappings = mappings.getMappings(FieldMapping.DIRECT_CHILD_ELT);
				FieldMapping[] attrMappings = mappings.getMappings(FieldMapping.ATTRIBUTE);
				boolean hasContentMappings = contentMappings.length > 0;
		//		boolean hasNodeMappings = (childMappings.size() + collectedMappings.size()) > 0;
				
				// TODO: if object has content mappings and no node mapping (must 
				// lookup --> Object.class), declare as complexType with simpleContent (text)
				
				// --- 1: declare type
				Element typeElt =  iXmlDoc.createElement("xs:complexType");
				schemaElt.appendChild(typeElt);
				typeElt.setAttribute("name", "_"+elements[i].getTagName());
				if(isAbstract)
					typeElt.setAttribute("abstract", "true");
				
				if(hasContentMappings)
					typeElt.setAttribute("mixed", "true");
				
				// --- add documentation
// not on type: on element				addComponentDoc(typeElt, elements[i]);

				// --- declare the inheritance
				ISchemaElement superType = elements[i].getSuperType();
				if(superType != null)
				{
					Element cc = iXmlDoc.createElement("xs:complexContent");
					typeElt.appendChild(cc);
					
					Element ext = iXmlDoc.createElement("xs:extension");
					// --- 'base' attribute points to a type
					ext.setAttribute("base", getPrefixAndTag(superType, true));
					cc.appendChild(ext);
					
					typeElt = ext;
				}
				
				Element sequence = null;
				
				// --- 1: child nodes
				for(int j=0; j<childMappings.length; j++)
				{
					FieldMapping fieldMapping = childMappings[j];
					if(sequence == null)
					{
						sequence = iXmlDoc.createElement("xs:sequence");
						typeElt.appendChild(sequence);
					}
					Element eltNode = iXmlDoc.createElement("xs:element");
					eltNode.setAttribute("name", fieldMapping.getName());
					eltNode.setAttribute("minOccurs", fieldMapping.isUseRequired() ? "1" : "0");
					eltNode.setAttribute("maxOccurs", "1");
					sequence.appendChild(eltNode);
					
					// --- add documentation
					addAttributeDoc(eltNode, fieldMapping);
					
					// --- declare type
					String simpleTypeName = getSimpleTypeName(fieldMapping.getBaseClass());
					if(simpleTypeName != null)
					{
						// --- this is a simple type
						if(fieldMapping.getMaxOccurs() > 1)
							// --- array not allowed for simple type ?
							System.out.println("Array of simple type not supported: '"+fieldMapping.getName()+"' ("+fieldMapping.getBaseClass()+"[])");
						else
							eltNode.setAttribute("type", simpleTypeName);
					}
					else
					{
						// --- this is a complex type
						Element ctNode = iXmlDoc.createElement("xs:complexType");
						eltNode.appendChild(ctNode);
						
						Element seqNode = iXmlDoc.createElement("xs:sequence");
						ctNode.appendChild(seqNode);
						
						Element eltChild = iXmlDoc.createElement("xs:element");
						ISchemaElement ref = _manager.getElement(fieldMapping.getBaseClass());
						if(ref.getType() == ISchemaElement.SIMPLE_TYPE_TYPE)
							// --- 'type' attribute points to the simple type
							eltChild.setAttribute("type", getPrefixAndTag(ref, false));
						else
							// --- 'ref' attribute points to the element
							eltChild.setAttribute("ref", getPrefixAndTag(ref, false));
// no						eltChild.setAttribute("minOccurs", String.valueOf(Math.max(1, fieldMapping.getMinOccurs())));
						eltChild.setAttribute("minOccurs", String.valueOf(fieldMapping.getMinOccurs()));
						eltChild.setAttribute("maxOccurs", fieldMapping.getMaxOccurs() == Integer.MAX_VALUE ? "unbounded" : String.valueOf(fieldMapping.getMaxOccurs()));
						seqNode.appendChild(eltChild);
					}
				}
				
				// --- 2: collected nodes
				for(int j=0; j<collectedMappings.length; j++)
				{
					FieldMapping fieldMapping = (FieldMapping)collectedMappings[j];
					if(sequence == null)
					{
						sequence = iXmlDoc.createElement("xs:sequence");
						typeElt.appendChild(sequence);
					}
					Element eltChild = iXmlDoc.createElement("xs:element");
					// --- 'ref' attribute points to the element
					ISchemaElement ref = _manager.getElement(fieldMapping.getBaseClass());
					if(ref.getType() == ISchemaElement.SIMPLE_TYPE_TYPE)
						// --- 'type' attribute points to the simple type
						eltChild.setAttribute("type", getPrefixAndTag(ref, false));
					else
						// --- 'ref' attribute points to the element
						eltChild.setAttribute("ref", getPrefixAndTag(ref, false));
					eltChild.setAttribute("minOccurs", String.valueOf(fieldMapping.getMinOccurs()));
					eltChild.setAttribute("maxOccurs", fieldMapping.getMaxOccurs() == Integer.MAX_VALUE ? "unbounded" : String.valueOf(fieldMapping.getMaxOccurs()));
					sequence.appendChild(eltChild);
					
					// --- add documentation
					addAttributeDoc(eltChild, fieldMapping);
				}
				
				if(hasContentMappings && sequence == null)
				{
					// --- allows a sequence on any type
					sequence = iXmlDoc.createElement("xs:sequence");
					typeElt.appendChild(sequence);
					Element any = iXmlDoc.createElement("xs:any");
					any.setAttribute("minOccurs", "0");
					any.setAttribute("maxOccurs", "unbounded");
					sequence.appendChild(any);
				}
				
				// --- 3: attributes
				for(int j=0; j<attrMappings.length; j++)
				{
					FieldMapping fieldMapping = (FieldMapping)attrMappings[j];
					Element attr = iXmlDoc.createElement("xs:attribute");
					// --- name
					attr.setAttribute("name", fieldMapping.getName());
					// --- use (required / optional)
					attr.setAttribute("use", fieldMapping.isUseRequired() ? "required" : "optional");
					typeElt.appendChild(attr);
					
					// --- type
					String simpleTypeName = getSimpleTypeName(fieldMapping.getBaseClass());
					if(simpleTypeName != null)
					{
						// --- this is a simple type
						if(fieldMapping.getMaxOccurs() > 1)
							// --- array not allowed for simple type ?
							System.out.println("Array of simple type not supported: '"+fieldMapping.getName()+"' ("+fieldMapping.getBaseClass()+"[])");
						else
							attr.setAttribute("type", simpleTypeName);
					}
					else
					{
//						throw new XmlMappingException("Unsupported attribute type: '"+fieldMapping.getType()+"'");
						System.out.println("Unsupported attribute type: '"+fieldMapping.getBaseClass()+"'");
					}
					addAttributeDoc(attr, fieldMapping);
				}
				
				// --- declare element
				Element elt = iXmlDoc.createElement("xs:element");
				elt.setAttribute("name", elements[i].getTagName());
				elt.setAttribute("type", "_"+elements[i].getTagName());
				if(isAbstract)
					elt.setAttribute("abstract", "true");
				
				// --- add documentation
				addComponentDoc(elt, elements[i]);
				
				// --- declare its substitution groups
				ISchemaElement[] substGroups = elements[i].getSubstitutionGroups();
				if(substGroups != null && substGroups.length > 0)
				{
					// --- 'substitutionGroup' attribute points to an element name
					if(substGroups.length > 1)
					{
						// --- more than one interfaces: not supported
						System.out.println("     + warning: class '"+elements[i].getMappedClass().getName()+"' has more than one interfaces:");
						for(int j=0; j<substGroups.length; j++)
							System.out.println("        ["+j+"] "+substGroups[j].getMappedClass().getName());
						System.out.println("     + Only the first one will be declared as substitution group: "+substGroups[0].getMappedClass().getName());
					}
					else
					{
						System.out.println("     + subst. group: "+substGroups[0].getMappedClass().getName());
					}
					elt.setAttribute("substitutionGroup", getPrefixAndTag(substGroups[0], false));
				}
				schemaElt.appendChild(elt);
			}
		}
	}
	private boolean addComponentDoc(Element iElement, ISchemaElement iElt)
	{
		// --- TODO: for all locales
//		String compName = MessagesHelper.getComponentName(Locale.getDefault(), iElt.getMappedClass());
		String doc = getMethod().getElementDetails(Locale.getDefault(), iElt.getMappedClass());
//		if(doc == null)
//			doc = MessagesHelper.getComponentName(Locale.getDefault(), iElt.getMappedClass());
		if(doc == null)
			return false;
		Element annotation = iElement.getOwnerDocument().createElement("xs:annotation");
		iElement.appendChild(annotation);
		Element documentation = iElement.getOwnerDocument().createElement("xs:documentation");
		annotation.appendChild(documentation);
//		documentation.setTextContent(doc);
		CDATASection cdata = iElement.getOwnerDocument().createCDATASection(doc);
		documentation.appendChild(cdata);
		return true;
	}
	private boolean addAttributeDoc(Element iElement, FieldMapping iMapping)
	{
		// --- TODO: for all locales
//		String compName = MessagesHelper.getComponentName(Locale.getDefault(), iElt.getMappedClass());
		String doc = getMethod().getMappingDetails(Locale.getDefault(), iMapping);
//		if(doc == null)
//			doc = MessagesHelper.getAttributeName(Locale.getDefault(), iElt.getMappedClass(), iAttributeName);
		if(doc == null)
			return false;
		Element annotation = iElement.getOwnerDocument().createElement("xs:annotation");
		iElement.appendChild(annotation);
		Element documentation = iElement.getOwnerDocument().createElement("xs:documentation");
		annotation.appendChild(documentation);
//		documentation.setTextContent(doc);
		CDATASection cdata = iElement.getOwnerDocument().createCDATASection(doc);
		documentation.appendChild(cdata);
		return true;
	}
	private boolean addEnumDoc(Element iElement, ISchemaElement iElt, String enulValueName)
	{
		// --- TODO: for all locales
//		String compName = MessagesHelper.getComponentName(Locale.getDefault(), iElt.getMappedClass());
		String doc = getMethod().getEnumItemDetails(Locale.getDefault(), iElt.getMappedClass(), enulValueName);
//		if(doc == null)
//			doc = MessagesHelper.getAttributeName(Locale.getDefault(), iElt.getMappedClass(), iAttributeName);
		if(doc == null)
			return false;
		Element annotation = iElement.getOwnerDocument().createElement("xs:annotation");
		iElement.appendChild(annotation);
		Element documentation = iElement.getOwnerDocument().createElement("xs:documentation");
		annotation.appendChild(documentation);
//		documentation.setTextContent(doc);
		CDATASection cdata = iElement.getOwnerDocument().createCDATASection(doc);
		documentation.appendChild(cdata);
		return true;
	}
	private String getSimpleTypeName(Class type)
	{
		String simpleTypeName = XsdHelper.getSimpleType(type);
		if(simpleTypeName != null)
			return "xs:"+simpleTypeName;
		ISchemaElement ref = _manager.getElement(type);
		if(ref == null)
		{
			return null;
		}
		else if(ref.getType() == ISchemaElement.ENUMERATION_TYPE)
		{
			// --- enumerated type
			return getPrefixAndTag(ref, false);
		}
		else if(ref.getType() == ISchemaElement.SIMPLE_TYPE_TYPE)
		{
			// --- simple type
			return getPrefixAndTag(ref, false);
		}
		return null;
	}
	private String getPrefixAndTag(ISchemaElement iElt, boolean iAsType)
	{
		if(iAsType && iElt.getType() == ISchemaElement.ELEMENT_TYPE)
		{
			// --- type name starts with '_' 
			if(iElt.getSchema() == this || iElt.getSchema().getName() == null)
				return "_"+iElt.getTagName();
			return iElt.getSchema().getName()+":_"+iElt.getTagName();
		}
		else
		{
			// --- enum or substitution group
			if(iElt.getSchema() == this || iElt.getSchema().getName() == null)
				return iElt.getTagName();
			return iElt.getSchema().getName()+":"+iElt.getTagName();
		}
	}
	private static String getPrefix(String iXsdFile)
	{
		String prefix = iXsdFile;
		int idx = prefix.lastIndexOf('/');
		if(idx >= 0)
			prefix = prefix.substring(idx+1);
		idx = prefix.indexOf('.');
		if(idx >= 0)
			prefix = prefix.substring(0, idx);
		return prefix;
	}
	/*
	public void buildFromXsd(Document iDoc)
	{
		Element schemaElt = iDoc.getDocumentElement();
		// --- read root package
		_package = schemaElt.getAttribute("targetNamespace");
		
		// --- read imports
		NodeList imports = schemaElt.getElementsByTagName("import");
		for(int i=0; i<imports.getLength(); i++)
		{
			if(imports.item(i) instanceof Element)
			{
				Element importElt = (Element)imports.item(i);
				String importPackage = importElt.getAttribute("namespace");
				String importLocation = importElt.getAttribute("schemaLocation");
				// --- look for prefix in root element attributes: xmlns:<prefix>="<package>"
				String importPrefix = null;
				NamedNodeMap attributes = schemaElt.getAttributes();
				for(int j=0; j<attributes.getLength(); j++)
				{
					if(attributes.item(j) instanceof Attr)
					{
						Attr attr = (Attr)attributes.item(j);
						if("xmlns".equals(attr.getPrefix()) && importPackage.equals(attr.getValue()))
						{
							// --- attribute found:
							importPrefix = attr.getLocalName();
							break;
						}
					}
				}
				if(importPrefix == null)
					// --- extract prefix from xsd filename
					importPrefix = getPrefix(importLocation);
				
				_manager.declareSchema(importPrefix, importLocation, importPackage);
			}
		}
		
		NodeList children = schemaElt.getChildNodes();
		for(int i=0; i<children.getLength(); i++)
		{
			if(children.item(i) instanceof Element)
			{
				Element node = (Element)children.item(i);
				if(node.getLocalName().equals("simpleType") || node.getLocalName().equals("element"))
				{
					// --- interface, class or enumerated type: retrieve class
                    try
                    {
	                    Class c = Class.forName(_package+"."+node.getAttribute("name"));
						getElement(c);
                    }
                    catch(ClassNotFoundException e)
                    {
                    }
				}
			}
		}
	}
	*/

	// =====================================================================
	// === SchemaElement
	// =====================================================================
	private static class SchemaElement implements ISchemaElement
	{
		private Class _class;
		private int _type;
		private String _tag;
		private ISchema _schema;
		private ISchemaElement[] _substitutionGroups;
		private ISchemaElement _superClass;
		
		public Class getMappedClass()
		{
		    return _class;
		}
		public ISchema getSchema()
		{
		    return _schema;
		}
		/*
		public XmlMappings getClassMappings()
		{
		    return _mappings;
		}
		*/
		public ISchemaElement[] getSubstitutionGroups()
		{
		    return _substitutionGroups;
		}
		public ISchemaElement getSuperType()
		{
		    return _superClass;
		}
		public String getTagName()
		{
		    return _tag;
		}
		public int getType()
		{
		    return _type;
		}
		public String toString()
		{
			return "SchemaElement["+(_type == ELEMENT_TYPE ? "ELT" : _type == ENUMERATION_TYPE ? "ENUM" : "INTERF")+":"+getMappedClass().getName()+"]";
		}
	}
}
